#!perl

use 5.010001;
use strict;
use warnings;

use Complete::Bash::History qw();
use Test::More 0.98;

# XXX test arg path
# XXX test arg max_hist_lines
# XXX test arg max_result

ok 1;

DONE_TESTING:
done_testing;
